@@ -7,14 +7,16 @@ from __future__ import absolute_import
 import select
 import threading
 import weakref
+import warnings
 from functools import wraps
 
+import imp
+
 from twisted.python import threadable
 from twisted.python.runtime import platform
 from twisted.python.failure import Failure
 from twisted.python.log import PythonLoggingObserver, err
 from twisted.internet.defer import maybeDeferred
-from twisted.internet import reactor
 from twisted.internet.task import LoopingCall
 
 from ._util import synchronized
@@ -105,7 +107,7 @@ class EventualResult(object):
     decorated with @run_in_reactor.
     """
 
-    def __init__(self, deferred, _reactor=reactor):
+    def __init__(self, deferred, _reactor):
         """
         The deferred parameter should be a Deferred or None indicating
         _connect_deferred will be called separately later.
@@ -180,6 +182,8 @@ class EventualResult(object):
         result.
         """
         if timeout is None:
+            warnings.warn("Unlimited timeouts are deprecated.",
+                          DeprecationWarning, stacklevel=3)
             # Queue.get(None) won't get interrupted by Ctrl-C...
             timeout = 2 ** 31
         self._result_set.wait(timeout)
@@ -207,6 +211,14 @@ class EventualResult(object):
             raise RuntimeError(
                 "EventualResult.wait() must not be run in the reactor thread.")
 
+        if imp.lock_held():
+            # If EventualResult.wait() is run during module import, if the
+            # Twisted code that is being run also imports something the result
+            # will be a deadlock. Even if that is not an issue it would
+            # prevent importing in other threads until the call returns.
+            raise RuntimeError(
+                "EventualResult.wait() must not be run at module import time.")
+
         result = self._result(timeout)
         if isinstance(result, Failure):
             result.raiseException()
@@ -286,18 +298,26 @@ class EventLoop(object):
     """
     Initialization infrastructure for running a reactor in a thread.
     """
-    def __init__(self, reactor, atexit_register,
+    def __init__(self, reactorFactory, atexit_register,
                  startLoggingWithObserver=None,
                  watchdog_thread=None,
                  reapAllProcesses=None):
-        self._reactor = reactor
+        """
+        reactorFactory: Zero-argument callable that returns a reactor.
+        atexit_register: atexit.register, or look-alike.
+        startLoggingWithObserver: Either None, or
+            twisted.python.log.startLoggingWithObserver or lookalike.
+        watchdog_thread: crochet._shutdown.Watchdog instance, or None.
+        reapAllProcesses: twisted.internet.process.reapAllProcesses or
+            lookalike.
+        """
+        self._reactorFactory = reactorFactory
         self._atexit_register = atexit_register
         self._startLoggingWithObserver = startLoggingWithObserver
         self._started = False
         self._lock = threading.Lock()
         self._watchdog_thread = watchdog_thread
         self._reapAllProcesses = reapAllProcesses
-        self._registry = ResultRegistry(self._reactor)
 
     def _startReapingProcesses(self):
         """
@@ -312,6 +332,8 @@ class EventLoop(object):
         The minimal amount of setup done by both setup() and no_setup().
         """
         self._started = True
+        self._reactor = self._reactorFactory()
+        self._registry = ResultRegistry(self._reactor)
         # We want to unblock EventualResult regardless of how the reactor is
         # run, so we always register this:
         self._reactor.addSystemEventTrigger(
@@ -335,11 +357,20 @@ class EventLoop(object):
             self._reactor.callFromThread(self._startReapingProcesses)
         if self._startLoggingWithObserver:
             observer = ThreadLogObserver(PythonLoggingObserver().emit)
-            self._reactor.callFromThread(
-                self._startLoggingWithObserver, observer, False)
+            def start():
+                # Twisted is going to override warnings.showwarning; let's
+                # make sure that has no effect:
+                from twisted.python import log
+                original = log.showwarning
+                log.showwarning = warnings.showwarning
+                self._startLoggingWithObserver(observer, False)
+                log.showwarning = original
+            self._reactor.callFromThread(start)
+
             # We only want to stop the logging thread once the reactor has
             # shut down:
-            self._reactor.addSystemEventTrigger("after", "shutdown", observer.stop)
+            self._reactor.addSystemEventTrigger("after", "shutdown",
+                                                observer.stop)
         t = threading.Thread(
             target=lambda: self._reactor.run(installSignalHandlers=False),
             name="CrochetReactor")
@@ -384,22 +415,49 @@ class EventLoop(object):
             self._registry.register(result)
             self._reactor.callFromThread(runs_in_reactor, result, args, kwargs)
             return result
+        wrapper.wrapped_function = function
         return wrapper
 
     def wait_for_reactor(self, function):
         """
+        DEPRECATED, use wait_for(timeout) instead.
+
         A decorator that ensures the wrapped function runs in the reactor thread.
 
         When the wrapped function is called, its result is returned or its
         exception raised. Deferreds are handled transparently.
         """
-        @wraps(function)
-        def wrapper(*args, **kwargs):
-            @self.run_in_reactor
-            def run():
-                return function(*args, **kwargs)
-            return run().wait()
-        return wrapper
+        warnings.warn("@wait_for_reactor is deprecated, use @wait_for instead",
+                      DeprecationWarning, stacklevel=2)
+        # This will timeout, in theory. In practice the process will be dead
+        # long before that.
+        return self.wait_for(2 ** 31)(function)
+
+    def wait_for(self, timeout):
+        """
+        A decorator factory that ensures the wrapped function runs in the
+        reactor thread.
+
+        When the wrapped function is called, its result is returned or its
+        exception raised. Deferreds are handled transparently. Calls will
+        timeout after the given number of seconds (a float), raising a
+        crochet.TimeoutError, and cancelling the Deferred being waited on.
+        """
+        def decorator(function):
+            @wraps(function)
+            def wrapper(*args, **kwargs):
+                @self.run_in_reactor
+                def run():
+                    return function(*args, **kwargs)
+                eventual_result = run()
+                try:
+                    return eventual_result.wait(timeout)
+                except TimeoutError:
+                    eventual_result.cancel()
+                    raise
+            wrapper.wrapped_function = function
+            return wrapper
+        return decorator
 
     def in_reactor(self, function):
         """
@@ -412,7 +470,6 @@ class EventLoop(object):
 
         When the wrapped function is called, an EventualResult is returned.
         """
-        import warnings
         warnings.warn("@in_reactor is deprecated, use @run_in_reactor",
                       DeprecationWarning, stacklevel=2)
         @self.run_in_reactor
