@@ -10,8 +10,10 @@ import time
 import gc
 import sys
 import weakref
+import tempfile
+import os
 
-from twisted.trial.unittest import TestCase, SkipTest
+from twisted.trial.unittest import TestCase
 from twisted.internet.defer import succeed, Deferred, fail, CancelledError
 from twisted.python.failure import Failure
 from twisted.python import threadable
@@ -34,7 +36,7 @@ from .._eventloop import (EventLoop, EventualResult, TimeoutError,
                           ResultRegistry, ReactorStopped)
 from .test_setup import FakeReactor
 from .. import (_main, setup, in_reactor, retrieve_result, _store, no_setup,
-                run_in_reactor, wait_for_reactor)
+                run_in_reactor, wait_for_reactor, wait_for)
 from ..tests import crochet_directory
 
 
@@ -48,7 +50,7 @@ class ResultRegistryTests(TestCase):
         ReactorStopped.
         """
         registry = ResultRegistry(FakeReactor())
-        er = EventualResult(None)
+        er = EventualResult(None, None)
         registry.register(er)
         registry.stop()
         self.assertRaises(ReactorStopped, er.wait, timeout=0)
@@ -59,7 +61,7 @@ class ResultRegistryTests(TestCase):
         raise ReactorStopped.
         """
         registry = ResultRegistry(FakeReactor())
-        er = EventualResult(None)
+        er = EventualResult(None, None)
         registry.stop()
         self.assertRaises(ReactorStopped, registry.register, er)
 
@@ -69,7 +71,7 @@ class ResultRegistryTests(TestCase):
         which already have a result.
         """
         registry = ResultRegistry(FakeReactor())
-        er = EventualResult(succeed(123))
+        er = EventualResult(succeed(123), None)
         registry.register(er)
         registry.stop()
         self.assertEqual(er.wait(), 123)
@@ -82,7 +84,7 @@ class ResultRegistryTests(TestCase):
         it from being garbage collected.
         """
         registry = ResultRegistry(FakeReactor())
-        er = EventualResult(None)
+        er = EventualResult(None, None)
         registry.register(er)
         ref = weakref.ref(er)
         del er
@@ -110,7 +112,7 @@ class EventualResultTests(TestCase):
         """
         wait() returns the value the Deferred fired with.
         """
-        dr = EventualResult(succeed(123))
+        dr = EventualResult(succeed(123), None)
         self.assertEqual(dr.wait(), 123)
 
     def test_later_success_result(self):
@@ -123,14 +125,14 @@ class EventualResultTests(TestCase):
             import time; time.sleep(0.01)
             d.callback(345)
         threading.Thread(target=fireSoon).start()
-        dr = EventualResult(d)
+        dr = EventualResult(d, None)
         self.assertEqual(dr.wait(), 345)
 
     def test_success_result_twice(self):
         """
         A second call to wait() returns same value as the first call.
         """
-        dr = EventualResult(succeed(123))
+        dr = EventualResult(succeed(123), None)
         self.assertEqual(dr.wait(), 123)
         self.assertEqual(dr.wait(), 123)
 
@@ -138,7 +140,7 @@ class EventualResultTests(TestCase):
         """
         wait() raises the exception the Deferred fired with.
         """
-        dr = EventualResult(fail(RuntimeError()))
+        dr = EventualResult(fail(RuntimeError()), None)
         self.assertRaises(RuntimeError, dr.wait)
 
     def test_later_failure_result(self):
@@ -151,14 +153,14 @@ class EventualResultTests(TestCase):
             time.sleep(0.01)
             d.errback(RuntimeError())
         threading.Thread(target=fireSoon).start()
-        dr = EventualResult(d)
+        dr = EventualResult(d, None)
         self.assertRaises(RuntimeError, dr.wait)
 
     def test_failure_result_twice(self):
         """
         A second call to wait() raises same value as the first call.
         """
-        dr = EventualResult(fail(ZeroDivisionError()))
+        dr = EventualResult(fail(ZeroDivisionError()), None)
         self.assertRaises(ZeroDivisionError, dr.wait)
         self.assertRaises(ZeroDivisionError, dr.wait)
 
@@ -167,7 +169,7 @@ class EventualResultTests(TestCase):
         If no result is available, wait(timeout) will throw a TimeoutError.
         """
         start = time.time()
-        dr = EventualResult(Deferred())
+        dr = EventualResult(Deferred(), None)
         self.assertRaises(TimeoutError, dr.wait, timeout=0.03)
         self.assertTrue(abs(time.time() - start - 0.03) < 0.005)
 
@@ -176,7 +178,7 @@ class EventualResultTests(TestCase):
         If no result is available, a second call to wait(timeout) will also
         result in a TimeoutError exception.
         """
-        dr = EventualResult(Deferred())
+        dr = EventualResult(Deferred(), None)
         self.assertRaises(TimeoutError, dr.wait, timeout=0.01)
         self.assertRaises(TimeoutError, dr.wait, timeout=0.01)
 
@@ -186,7 +188,7 @@ class EventualResultTests(TestCase):
         wait() will return it.
         """
         d = Deferred()
-        dr = EventualResult(d)
+        dr = EventualResult(d, None)
         self.assertRaises(TimeoutError, dr.wait, timeout=0.01)
         d.callback(u"value")
         self.assertEqual(dr.wait(), u"value")
@@ -198,7 +200,7 @@ class EventualResultTests(TestCase):
         """
         self.patch(threadable, "isInIOThread", lambda: True)
         d = Deferred()
-        dr = EventualResult(d)
+        dr = EventualResult(d, None)
         self.assertRaises(RuntimeError, dr.wait, 0)
 
     def test_cancel(self):
@@ -222,7 +224,7 @@ class EventualResultTests(TestCase):
         """
         EventualResult.stash() stores the object in the global ResultStore.
         """
-        dr = EventualResult(Deferred())
+        dr = EventualResult(Deferred(), None)
         uid = dr.stash()
         self.assertIdentical(dr, _store.retrieve(uid))
 
@@ -235,21 +237,21 @@ class EventualResultTests(TestCase):
             1/0
         except:
             f = Failure()
-        dr = EventualResult(fail(f))
+        dr = EventualResult(fail(f), None)
         self.assertIdentical(dr.original_failure(), f)
 
     def test_original_failure_no_result(self):
         """
         If there is no result yet, original_failure() returns None.
         """
-        dr = EventualResult(Deferred())
+        dr = EventualResult(Deferred(), None)
         self.assertIdentical(dr.original_failure(), None)
 
     def test_original_failure_not_error(self):
         """
         If the result is not an error, original_failure() returns None.
         """
-        dr = EventualResult(succeed(3))
+        dr = EventualResult(succeed(3), None)
         self.assertIdentical(dr.original_failure(), None)
 
     def test_error_logged_no_wait(self):
@@ -257,7 +259,7 @@ class EventualResultTests(TestCase):
         If the result is an error and wait() was never called, the error will
         be logged once the EventualResult is garbage-collected.
         """
-        dr = EventualResult(fail(ZeroDivisionError()))
+        dr = EventualResult(fail(ZeroDivisionError()), None)
         del dr
         gc.collect()
         excs = self.flushLoggedErrors(ZeroDivisionError)
@@ -269,7 +271,7 @@ class EventualResultTests(TestCase):
         error will be logged once the EventualResult is garbage-collected.
         """
         d = Deferred()
-        dr = EventualResult(d)
+        dr = EventualResult(d, None)
         try:
             dr.wait(0)
         except TimeoutError:
@@ -288,7 +290,7 @@ class EventualResultTests(TestCase):
         EventualResult are lost, the error is still logged.
         """
         d = Deferred()
-        dr = EventualResult(d)
+        dr = EventualResult(d, None)
         del dr
         d.errback(ZeroDivisionError())
         gc.collect()
@@ -327,7 +329,7 @@ t.setDaemon(True)
 t.start()
 
 d = Deferred()
-e = crochet.EventualResult(d)
+e = crochet.EventualResult(d, None)
 
 try:
     # Queue.get() has special non-interruptible behavior if not given timeout,
@@ -337,7 +339,7 @@ except KeyboardInterrupt:
     sys.exit(23)
 """
         kw = { 'cwd': crochet_directory }
-        # on Windows the only way to interrupt a subprocess reliably is to 
+        # on Windows the only way to interrupt a subprocess reliably is to
         # create a new process group:
         # http://docs.python.org/2/library/subprocess.html#subprocess.CREATE_NEW_PROCESS_GROUP
         if platform.type.startswith('win'):
@@ -345,20 +347,13 @@ except KeyboardInterrupt:
         process = subprocess.Popen([sys.executable, "-c", program], **kw)
         self.assertEqual(process.wait(), 23)
 
-    def test_reactor_stop_unblocks_EventualResult(self):
-        """
-        Any EventualResult.wait() calls still waiting when the reactor has
-        stopped will get a ReactorStopped exception.
-        """
-        raise SkipTest("Not done yet.")
-
     def test_connect_deferred(self):
         """
         If an EventualResult is created with None,
         EventualResult._connect_deferred can be called later to register a
         Deferred as the one it is wrapping.
         """
-        er = EventualResult(None)
+        er = EventualResult(None, None)
         self.assertRaises(TimeoutError, er.wait, 0)
         d = Deferred()
         er._connect_deferred(d)
@@ -462,6 +457,34 @@ else:
                                    cwd=crochet_directory,)
         self.assertEqual(process.wait(), 23)
 
+    def test_noWaitingDuringImport(self):
+        """
+        EventualResult.wait() raises an exception if called while a module is
+        being imported.
+
+        This prevents the imports from taking a long time, preventing other
+        imports from running in other threads. It also prevents deadlocks,
+        which can happen if the code being waited on also tries to import
+        something.
+        """
+        if sys.version_info[0] > 2:
+            from unittest import SkipTest
+            raise SkipTest("This test is too fragile (and insufficient) on "
+                           "Python 3 - see "
+                           "https://github.com/itamarst/crochet/issues/43")
+        directory = tempfile.mktemp()
+        os.mkdir(directory)
+        sys.path.append(directory)
+        self.addCleanup(sys.path.remove, directory)
+        with open(os.path.join(directory, "shouldbeunimportable.py"), "w") as f:
+            f.write("""\
+from crochet import EventualResult
+from twisted.internet.defer import Deferred
+
+EventualResult(Deferred(), None).wait(1.0)
+""")
+        self.assertRaises(RuntimeError, __import__, "shouldbeunimportable")
+
 
 class InReactorTests(TestCase):
     """
@@ -473,7 +496,7 @@ class InReactorTests(TestCase):
         The function decorated with in_reactor has the same name as the
         original function.
         """
-        c = EventLoop(FakeReactor(), lambda f, g: None)
+        c = EventLoop(lambda: FakeReactor(), lambda f, g: None)
 
         @c.in_reactor
         def some_name(reactor):
@@ -486,7 +509,9 @@ class InReactorTests(TestCase):
         thread, and takes the reactor as its first argument.
         """
         myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
+        c = EventLoop(lambda: myreactor, lambda f, g: None)
+        c.no_setup()
+
         calls = []
 
         @c.in_reactor
@@ -513,7 +538,8 @@ class InReactorTests(TestCase):
             return wrapper
 
         myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
+        c = EventLoop(lambda: myreactor, lambda f, g: None)
+        c.no_setup()
         c.run_in_reactor = fake_run_in_reactor
 
 
@@ -536,7 +562,7 @@ class RunInReactorTests(TestCase):
         The function decorated with run_in_reactor has the same name as the
         original function.
         """
-        c = EventLoop(FakeReactor(), lambda f, g: None)
+        c = EventLoop(lambda: FakeReactor(), lambda f, g: None)
 
         @c.run_in_reactor
         def some_name():
@@ -549,7 +575,8 @@ class RunInReactorTests(TestCase):
         thread.
         """
         myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
+        c = EventLoop(lambda: myreactor, lambda f, g: None)
+        c.no_setup()
         calls = []
 
         @c.run_in_reactor
@@ -566,7 +593,8 @@ class RunInReactorTests(TestCase):
         argument.
         """
         myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
+        c = EventLoop(lambda: myreactor, lambda f, g: None)
+        c.no_setup()
 
         @c.run_in_reactor
         def passthrough(argument):
@@ -610,7 +638,8 @@ class RunInReactorTests(TestCase):
         EventualResult hooked up to a Deferred wrapping the exception.
         """
         myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
+        c = EventLoop(lambda: myreactor, lambda f, g: None)
+        c.no_setup()
 
         @c.run_in_reactor
         def raiser():
@@ -625,7 +654,8 @@ class RunInReactorTests(TestCase):
         @run_in_reactor registers the EventualResult in the ResultRegistry.
         """
         myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
+        c = EventLoop(lambda: myreactor, lambda f, g: None)
+        c.no_setup()
 
         @c.run_in_reactor
         def run():
@@ -634,65 +664,110 @@ class RunInReactorTests(TestCase):
         result = run()
         self.assertIn(result, c._registry._results)
 
+    def test_wrapped_function(self):
+        """
+        The function wrapped by @run_in_reactor can be accessed via the
+        `wrapped_function` attribute.
+        """
+        c = EventLoop(lambda: None, lambda f, g: None)
+        def func():
+            pass
+        wrapper = c.run_in_reactor(func)
+        self.assertIdentical(wrapper.wrapped_function, func)
 
-class WaitForReactorTests(TestCase):
+
+class WaitTestsMixin(object):
     """
-    Tests for the wait_for_reactor decorator.
+    Tests mixin for the wait_for_reactor/wait_for decorators.
     """
+    def setUp(self):
+        self.reactor = FakeReactor()
+        self.eventloop = EventLoop(lambda: self.reactor, lambda f, g: None)
+        self.eventloop.no_setup()
+
+    def decorator(self):
+        """
+        Return a callable that decorates a function, using the decorator being
+        tested.
+        """
+        raise NotImplementedError()
+
+    def make_wrapped_function(self):
+        """
+        Return a function wrapped with the decorator being tested that returns
+        its first argument, or raises it if it's an exception.
+        """
+        decorator = self.decorator()
+        @decorator
+        def passthrough(argument):
+            if isinstance(argument, Exception):
+                raise argument
+            return argument
+        return passthrough
+
     def test_name(self):
         """
-        The function decorated with run_in_reactor has the same name as the
+        The function decorated with the wait decorator has the same name as the
         original function.
         """
-        c = EventLoop(FakeReactor(), lambda f, g: None)
-
-        @c.wait_for_reactor
-        def some_name():
+        decorator = self.decorator()
+        @decorator
+        def some_name(argument):
             pass
         self.assertEqual(some_name.__name__, "some_name")
 
+    def test_wrapped_function(self):
+        """
+        The function wrapped by the wait decorator can be accessed via the
+        `wrapped_function` attribute.
+        """
+        decorator = self.decorator()
+        def func():
+            pass
+        wrapper = decorator(func)
+        self.assertIdentical(wrapper.wrapped_function, func)
+
     def test_reactor_thread_disallowed(self):
         """
-        Functions decorated with wait_for_reactor() cannot be called from the
+        Functions decorated with the wait decorator cannot be called from the
         reactor thread.
         """
         self.patch(threadable, "isInIOThread", lambda: True)
-        c = EventLoop(FakeReactor(), lambda f, g: None)
-        @c.wait_for_reactor
-        def f():
-            pass
-        self.assertRaises(RuntimeError, f)
+        f = self.make_wrapped_function()
+        self.assertRaises(RuntimeError, f, None)
 
     def test_wait_for_reactor_thread(self):
         """
-        The function decorated with wait_for_reactor is run in the reactor
+        The function decorated with the wait decorator is run in the reactor
         thread.
         """
-        myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
-        calls = []
+        in_call_from_thread = []
+        decorator = self.decorator()
 
-        @c.wait_for_reactor
+        @decorator
+        def func():
+            in_call_from_thread.append(self.reactor.in_call_from_thread)
+
+        in_call_from_thread.append(self.reactor.in_call_from_thread)
+        func()
+        in_call_from_thread.append(self.reactor.in_call_from_thread)
+        self.assertEqual(in_call_from_thread, [False, True, False])
+
+    def test_arguments(self):
+        """
+        The function decorated with wait decorator gets all arguments passed
+        to the wrapper.
+        """
+        calls = []
+        decorator = self.decorator()
+
+        @decorator
         def func(a, b, c):
-            self.assertTrue(myreactor.in_call_from_thread)
             calls.append((a, b, c))
 
         func(1, 2, c=3)
         self.assertEqual(calls, [(1, 2, 3)])
 
-    def make_wrapped_function(self):
-        """
-        Return a function wrapped with wait_for_reactor that returns its first
-        argument.
-        """
-        myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
-
-        @c.wait_for_reactor
-        def passthrough(argument):
-            return argument
-        return passthrough
-
     def test_deferred_success_result(self):
         """
         If the underlying function returns a Deferred, the wrapper returns a
@@ -725,14 +800,8 @@ class WaitForReactorTests(TestCase):
         If the underlying function throws an exception, the wrapper raises
         that exception.
         """
-        myreactor = FakeReactor()
-        c = EventLoop(myreactor, lambda f, g: None)
-
-        @c.wait_for_reactor
-        def raiser():
-            1/0
-
-        self.assertRaises(ZeroDivisionError, raiser)
+        raiser = self.make_wrapped_function()
+        self.assertRaises(ZeroDivisionError, raiser, ZeroDivisionError())
 
     def test_control_c_is_possible(self):
         """
@@ -765,7 +834,7 @@ t = threading.Thread(target=interrupt)
 t.setDaemon(True)
 t.start()
 
-@crochet.wait_for_reactor
+@crochet.%s
 def wait():
     return Deferred()
 
@@ -773,7 +842,7 @@ try:
     wait()
 except KeyboardInterrupt:
     sys.exit(23)
-"""
+""" % (self.DECORATOR_CALL,)
         kw = { 'cwd': crochet_directory }
         if platform.type.startswith('win'):
             kw['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP
@@ -794,7 +863,7 @@ from twisted.internet import reactor
 import crochet
 crochet.setup()
 
-@crochet.wait_for_reactor
+@crochet.%s
 def run():
     reactor.callLater(0.1, reactor.stop)
     return Deferred()
@@ -803,12 +872,60 @@ try:
     er = run()
 except crochet.ReactorStopped:
     sys.exit(23)
-"""
+"""  % (self.DECORATOR_CALL,)
         process = subprocess.Popen([sys.executable, "-c", program],
                                    cwd=crochet_directory)
         self.assertEqual(process.wait(), 23)
 
 
+class WaitForReactorTests(WaitTestsMixin, TestCase):
+    """
+    Tests for the wait_for_reactor decorator.
+    """
+    DECORATOR_CALL = "wait_for_reactor"
+
+    def decorator(self):
+        return self.eventloop.wait_for_reactor
+
+
+class WaitForTests(WaitTestsMixin, TestCase):
+    """
+    Tests for the wait_for_reactor decorator.
+    """
+    DECORATOR_CALL = "wait_for(timeout=5)"
+
+    def decorator(self):
+        return lambda func: self.eventloop.wait_for(timeout=5)(func)
+
+    def test_timeoutRaises(self):
+        """
+        If a function wrapped with wait_for hits the timeout, it raises
+        TimeoutError.
+        """
+        @self.eventloop.wait_for(timeout=0.5)
+        def times_out():
+            return Deferred().addErrback(lambda f: f.trap(CancelledError))
+
+        start = time.time()
+        self.assertRaises(TimeoutError, times_out)
+        self.assertTrue(abs(time.time() - start - 0.5) < 0.1)
+
+    def test_timeoutCancels(self):
+        """
+        If a function wrapped with wait_for hits the timeout, it cancels
+        the underlying Deferred.
+        """
+        result = Deferred()
+        error = []
+        result.addErrback(error.append)
+
+        @self.eventloop.wait_for(timeout=0.0)
+        def times_out():
+            return result
+        self.assertRaises(TimeoutError, times_out)
+        self.assertIsInstance(error[0].value, CancelledError)
+
+
 class PublicAPITests(TestCase):
     """
     Tests for the public API.
@@ -818,7 +935,7 @@ class PublicAPITests(TestCase):
         Creating an EventLoop object, as is done in crochet.__init__, does not
         call any methods on the objects it is created with.
         """
-        c = EventLoop(None, lambda f, g: 1/0, lambda *args: 1/0,
+        c = EventLoop(lambda: None, lambda f, g: 1/0, lambda *args: 1/0,
                       watchdog_thread=object(), reapAllProcesses=lambda: 1/0)
         del c
 
@@ -827,7 +944,6 @@ class PublicAPITests(TestCase):
         An EventLoop object configured with the real reactor and
         _shutdown.register is exposed via its public methods.
         """
-        from twisted.internet import reactor
         from twisted.python.log import startLoggingWithObserver
         from crochet import _shutdown
         self.assertIsInstance(_main, EventLoop)
@@ -836,16 +952,25 @@ class PublicAPITests(TestCase):
         self.assertEqual(_main.in_reactor, in_reactor)
         self.assertEqual(_main.run_in_reactor, run_in_reactor)
         self.assertEqual(_main.wait_for_reactor, wait_for_reactor)
-        self.assertIdentical(_main._reactor, reactor)
+        self.assertEqual(_main.wait_for, wait_for)
         self.assertIdentical(_main._atexit_register, _shutdown.register)
-        self.assertIdentical(_main._startLoggingWithObserver, startLoggingWithObserver)
+        self.assertIdentical(_main._startLoggingWithObserver,
+                             startLoggingWithObserver)
         self.assertIdentical(_main._watchdog_thread, _shutdown._watchdog)
 
+    def test_eventloop_api_reactor(self):
+        """
+        The publicly exposed EventLoop will, when setup, use the global reactor.
+        """
+        from twisted.internet import reactor
+        _main.no_setup()
+        self.assertIdentical(_main._reactor, reactor)
+
     def test_retrieve_result(self):
         """
         retrieve_result() calls retrieve() on the global ResultStore.
         """
-        dr = EventualResult(Deferred())
+        dr = EventualResult(Deferred(), None)
         uid = dr.stash()
         self.assertIdentical(dr, retrieve_result(uid))
 
